#!/usr/bin/env python
"""
Convert a pdf to an image epub
"""

import os, sys, logging
import subprocess
import tempfile
import shutil

DefaultWidth = 550
DefaultHeight = 747

__version__ = 0.3
__author__ = 'Alejandro Dubrovsky <alito@organicrobot.com>'

def getLevelFromVerbosity(verbosity):
	"""
	Mapping of verbosity to log level
	"""
	loggingLevel = logging.ERROR
	if isinstance(verbosity, bool) and verbosity:
		loggingLevel = logging.DEBUG
	elif verbosity == 1:
		loggingLevel = logging.WARNING
	elif verbosity == 2:
		loggingLevel = logging.INFO
	elif verbosity >= 3:
		loggingLevel = logging.DEBUG

	return loggingLevel


def pdfToHtml(filename, directory):
	command = 'pdftohtml -enc UTF-8 -nodrm -nomerge -p -noframes'.split()
	command.append(filename)
	command.append(os.path.join(directory,'index.html'))
	logging.info("Calling %s" % " ".join(command))
	code = subprocess.call(command)
	if code != 0:
		logging.error("Failed to convert %s. Code: %s" % (filename, code))
	return code == 0
				   

def getTemplateDirectory():
	"""
	Return relative path to where the template files are
	"""
	
	current = __file__
	tree, filename = os.path.split(current)
	return os.path.join(tree, 'template')


def getTitle(filename):
	tree, filePart = os.path.split(filename)
	title, extension = os.path.splitext(filePart)
	return title


def resizeImages(directory, width, height):
	"""
	Split images into four and resize them to target size
	"""
	import math
	try:
		from PIL import Image
	except ImportError:
		print >> sys.stderr, "You don't have PIL installed.  Get it from http://www.pythonware.com/products/pil/"
		sys.exit(2)
	
	images = getImages(directory)

	targetSize = '%sx%s' % (width, height)
	for image in images:
		fullPath = os.path.join(directory, image) 
		i = Image.open(fullPath)
		width, height = i.size
		halfWidth = int(math.ceil(width / 2.0))
		halfHeight = int(math.ceil(height / 2.0))
		baseName, extension = os.path.splitext(image)
		geometry = "%sx%s" % (halfWidth, halfHeight)
		for index, offset in enumerate(['+0+0','+0+%s' % halfHeight, '+%s+0' % halfWidth, '+%s+%s' % (halfWidth, halfHeight)]):
			fullGeometry = "%s%s" % (geometry, offset)
			command = 'convert -crop'.split()
			command.append(fullGeometry)
			command.append('+repage')
			command.extend(['-resize',targetSize])
			command.extend([fullPath, os.path.join(directory, '%s_%s%s' % (baseName,index,extension))])
			logging.info("Calling %s" % " ".join(command))
			code = subprocess.call(command)
			if code != 0:
				logging.error("Error converting")
				return False
		os.remove(fullPath)

	return True
			

def getImages(directory):
	"""
	Get list of images in directory
	"""

	def giveItANumber(filename):
		filename = "".join(f for f in filename if f in '0123456789_')
		bits = filename.split('_')
		value = 0
		multiplier = 1
		for bit in reversed(bits):
			try:
				current = int(bit)
				value += multiplier * current
				multiplier *= 1000
			except ValueError:
				pass
		return value
				
	
	contents = os.listdir(directory)
	images = [content for content in contents if content.endswith('.png')]
	images.sort(key=giveItANumber)
	return images


def copyGenericBits(source, target):
	GenericFiles = ['mimetype','stylesheet.css']
	GenericDirectories = ['META-INF']

	success = True
	
	for filename in GenericFiles:
		logging.info("Copying %s" % filename)		
		shutil.copy(os.path.join(source, filename), target)

	for directory in GenericDirectories:
		logging.info("Copying %s" % directory)				
		try:
			shutil.copytree(os.path.join(source,directory), os.path.join(target,directory),symlinks=True)
		except OSError, e:
			success = False
			logging.error("Failed to copy %s to %s: %s" % (directory, target, e))
			
	return success


def copyTemplateBits(source, target, title, images):
	import uuid
	pageTemplateFilename = 'index.html'
	TemplatedFiles = ['toc.ncx','content.opf']

	substitutes = {}
	substitutes['title'] = title
	substitutes['pageCount'] = len(images)
	substitutes['uuid'] = str(uuid.uuid1())


	fin = open(os.path.join(source, pageTemplateFilename))
	pageContent = fin.read()
	fin.close()
	
	contentItemTemplate = '<item href="%(filename)s" id="idi%(id)s" media-type="image/png"/>'
	contentPageTemplate = '<item href="%(pageFilename)s" id="%(pageID)s" media-type="application/xhtml+xml"/>'

	opfContents = []
	pages = []
	for index, filename in enumerate(images):
		pageFilename = 'index%s.html' % (index + 1)
		pageID = 'id%s' % (index + 1)
		opfContents.append(contentItemTemplate % {'filename' : filename, 'id' : (index + 1)})
		opfContents.append(contentPageTemplate % {'pageFilename' : pageFilename, 'pageID' : pageID})

		fout = open(os.path.join(target, pageFilename),'w')
		fout.write(pageContent % {'title' : title, 'filename' : filename})
		fout.close()

		pages.append((pageFilename, pageID))
	
	substitutes['manifestPages'] = '\n'.join(opfContents)

	imageTags = []
	imageTemplate = ''

	
	for filename in images:
		imageTags.append(imageTemplate % {'filename' : filename})
	substitutes['images'] = '<br/>\n'.join(imageTags)
	substitutes['firstImage'] = images[0]

	spinePages = '\n'.join(('<itemref idref="%(pageID)s"/>' % {'pageID' : pageID}) for (filename, pageID) in pages)
	substitutes['spinePages'] = spinePages
	
	for filename in TemplatedFiles:
		logging.info("Creating %s" % filename)
		fin = open(os.path.join(source, filename))
		contents = fin.read()
		fin.close()

		fout = open(os.path.join(target,filename),'w')
		fout.write(contents % substitutes)
		fout.close()

		

def zipItUp(tempDir, name):
	currentDirectory = os.path.abspath(os.curdir)
	zipName = '%s.zip' % name
	try:
		os.chdir(tempDir)
		command = 'zip -r'.split()
		command.append(zipName)
		command.append(".")
		logging.info("Calling %s" % " ".join(command))
		code = subprocess.call(command)
		if code != 0:
			logging.error("Failed to zip. Code: %s" % code)
	finally:
		os.chdir(currentDirectory)

	return os.path.join(tempDir,zipName)


def main(args):
	from optparse import OptionParser
	
	parser = OptionParser('%prog [-v] [-h] [options] <pdf file>', epilog=__doc__)
	parser.add_option("-v", "--verbose", dest="verbosity", default=0, action="count", help="Verbosity.  Invoke many times for higher verbosity")
	parser.add_option("-o", "--output", dest="outputFilename", default=None,
					  help="Output filename.  If not specified, it is like the input with an epub extension")
	parser.add_option("--keep-work", dest="keepWork", default=False, action="store_true",
					  help="Leave temporary directory behind")
	parser.add_option("--width", dest="width", type="int", default=DefaultWidth,
					  help="Width of resultant images (default: %default)")	
	parser.add_option("--height", dest="height", type="int", default=DefaultHeight,
					  help="Height of resultant images (default: %default)")	
	
	options, rest = parser.parse_args(args)

	if not rest:
		parser.error("Need a file")
	elif len(rest) > 1:
		parser.error("At most one file")


	logging.basicConfig(level=getLevelFromVerbosity(options.verbosity))
		
	inputFilename = os.path.expanduser(rest[0])
	if not os.path.isfile(inputFilename):
		logging.error("%s not found" % rest[0])
		return 2

	title = getTitle(inputFilename)
	tempDir = tempfile.mkdtemp()
	logging.info("Using %s as a temporary directory" % tempDir)
	converted = pdfToHtml(inputFilename, tempDir)
	if not converted:
		return 2

	if not resizeImages(tempDir, options.width, options.height):
		return 2
	
	copyGenericBits(getTemplateDirectory(), tempDir)
	images = getImages(tempDir)

	copyTemplateBits(getTemplateDirectory(), tempDir, title, images)

	epub = zipItUp(tempDir, title)
	if options.outputFilename:
		outputFilename = options.outputFilename
	else:
		outputFilename = "%s.epub" % title

	shutil.move(epub, outputFilename)

	if not options.keepWork:
		shutil.rmtree(tempDir)
	
	return 0


if __name__ == '__main__':
	sys.exit(main(sys.argv[1:]))

